@@ -1,11 +1,12 @@
 """Notifications jobs."""
 import datetime
 import traceback
-from typing import Union
+from typing import Optional, List
+import uuid
 from django.conf import settings
-from circuit_maintenance_parser import ParsingError, init_provider
+from circuit_maintenance_parser import ProviderError, init_provider, NotificationData, Maintenance
 from nautobot.circuits.models import Circuit, Provider
-from nautobot.extras.jobs import Job
+from nautobot.extras.jobs import Job, BooleanVar
 from nautobot_circuit_maintenance.models import (
     CircuitImpact,
     CircuitMaintenance,
@@ -18,26 +19,26 @@ from .sources import get_notifications, MaintenanceNotification
 
 
 # pylint: disable=broad-except
-PLUGIN_SETTINGS = settings.PLUGINS_CONFIG["nautobot_circuit_maintenance"]
+PLUGIN_SETTINGS = settings.PLUGINS_CONFIG.get("nautobot_circuit_maintenance", {})
+MAX_INITIAL_DAYS_SINCE = 365
 
 
 def create_circuit_maintenance(
-    logger: Job, maintenance_id: str, parsed_notification: ParsedNotification
+    logger: Job, maintenance_id: str, parser_maintenance: Maintenance, provider: Provider
 ) -> CircuitMaintenance:
     """Handles the creation of a new circuit maintenance."""
-    provider = Provider.objects.filter(slug=parsed_notification.slug()).last()
-
-    circuit_maintenance_entry = CircuitMaintenance.objects.create(
+    circuit_maintenance_entry = CircuitMaintenance(
         name=maintenance_id,
-        start_time=datetime.datetime.fromtimestamp(parsed_notification.start),
-        end_time=datetime.datetime.fromtimestamp(parsed_notification.end),
-        description=parsed_notification.summary,
-        status=parsed_notification.status,
+        start_time=datetime.datetime.fromtimestamp(parser_maintenance.start),
+        end_time=datetime.datetime.fromtimestamp(parser_maintenance.end),
+        description=parser_maintenance.summary,
+        status=parser_maintenance.status,
     )
+    circuit_maintenance_entry.save()
     logger.log_success(obj=circuit_maintenance_entry, message="Created Circuit Maintenance.")
 
-    for circuit in parsed_notification.circuits:
-        circuit_entry = Circuit.objects.filter(cid=circuit.circuit_id, provider=provider.pk).last()
+    for circuit in parser_maintenance.circuits:
+        circuit_entry = Circuit.objects.filter(cid=circuit.circuit_id, provider=provider).last()
         if circuit_entry:
             circuit_impact_entry = CircuitImpact.objects.create(
                 maintenance=circuit_maintenance_entry,
@@ -66,7 +67,7 @@ def create_circuit_maintenance(
                 ),
             )
     if not CircuitImpact.objects.filter(maintenance=circuit_maintenance_entry):
-        logger.log_warning(message=f"Circuit Maintenance {maintenance_id} has no valid Circuit IDs")
+        logger.log_warning(message=f"Circuit Maintenance {maintenance_id} has none Circuit IDs in the DB.")
 
     return circuit_maintenance_entry
 
@@ -75,31 +76,30 @@ def update_circuit_maintenance(
     logger: Job,
     circuit_maintenance_entry: CircuitMaintenance,
     maintenance_id: str,
-    parsed_notification: ParsedNotification,
+    parser_maintenance: Maintenance,
+    provider: Provider,
 ):  # pylint: disable=too-many-locals
     """Handles the update of an existent circuit maintenance."""
-    provider = Provider.objects.filter(slug=parsed_notification.slug()).last()
-
-    circuit_maintenance_entry.description = parsed_notification.summary
-    circuit_maintenance_entry.status = parsed_notification.status
-    circuit_maintenance_entry.start_time = datetime.datetime.fromtimestamp(parsed_notification.start)
-    circuit_maintenance_entry.end_time = datetime.datetime.fromtimestamp(parsed_notification.end)
+    circuit_maintenance_entry.description = parser_maintenance.summary
+    circuit_maintenance_entry.status = parser_maintenance.status
+    circuit_maintenance_entry.start_time = datetime.datetime.fromtimestamp(parser_maintenance.start)
+    circuit_maintenance_entry.end_time = datetime.datetime.fromtimestamp(parser_maintenance.end)
     circuit_maintenance_entry.ack = False
     circuit_maintenance_entry.save()
 
     circuit_entries = CircuitImpact.objects.filter(maintenance=circuit_maintenance_entry)
 
-    parsed_cids = {parsed_circuit.circuit_id for parsed_circuit in parsed_notification.circuits}
+    new_cids = {parsed_circuit.circuit_id for parsed_circuit in parser_maintenance.circuits}
     existing_cids = {circuit_entry.circuit.cid for circuit_entry in circuit_entries}
 
-    cids_to_update = parsed_cids & existing_cids
-    cids_to_create = parsed_cids - existing_cids
-    cids_to_remove = existing_cids - parsed_cids
+    cids_to_update = new_cids & existing_cids
+    cids_to_create = new_cids - existing_cids
+    cids_to_remove = existing_cids - new_cids
 
     for cid in cids_to_create:
         circuit_entry = Circuit.objects.filter(cid=cid, provider=provider.pk).last()
         circuit = [
-            parsed_circuit for parsed_circuit in parsed_notification.circuits if parsed_circuit.circuit_id == cid
+            parsed_circuit for parsed_circuit in parser_maintenance.circuits if parsed_circuit.circuit_id == cid
         ][0]
         if circuit_entry:
             circuit_impact_entry = CircuitImpact.objects.create(
@@ -136,7 +136,7 @@ def update_circuit_maintenance(
             circuit=circuit_entry, maintenance=circuit_maintenance_entry
         ).last()
         circuit = [
-            parsed_circuit for parsed_circuit in parsed_notification.circuits if parsed_circuit.circuit_id == cid
+            parsed_circuit for parsed_circuit in parser_maintenance.circuits if parsed_circuit.circuit_id == cid
         ][0]
         circuitimpact_entry.impact = circuit.impact
         circuitimpact_entry.save()
@@ -148,74 +148,62 @@ def update_circuit_maintenance(
     logger.log_info(obj=circuit_maintenance_entry, message=f"Updated Circuit Maintenance {maintenance_id}")
 
 
-def process_parsed_notification(logger: Job, parsed_notification: ParsedNotification, raw_entry: RawNotification):
-    """Processes a Parsed Notification, creating or updating the related Circuit Maintenance."""
-    maintenance_id = str(parsed_notification.maintenance_id)
+def create_or_update_circuit_maintenance(
+    logger: Job, parser_maintenance: Maintenance, raw_entry: RawNotification, provider: Provider
+) -> CircuitMaintenance:
+    """Processes a Maintenance, creating or updating the related Circuit Maintenance.
+
+    It returns the CircuitMaintenance entry created or updated.
+    """
+    maintenance_id = f"{raw_entry.provider.slug}-{parser_maintenance.maintenance_id}"
     circuit_maintenance_entry = CircuitMaintenance.objects.filter(name=maintenance_id).last()
 
     if circuit_maintenance_entry:
-        update_circuit_maintenance(logger, circuit_maintenance_entry, maintenance_id, parsed_notification)
+        update_circuit_maintenance(logger, circuit_maintenance_entry, maintenance_id, parser_maintenance, provider)
     else:
-        circuit_maintenance_entry = create_circuit_maintenance(logger, maintenance_id, parsed_notification)
+        circuit_maintenance_entry = create_circuit_maintenance(logger, maintenance_id, parser_maintenance, provider)
 
-    # Insert parsed notification in DB
-    parsed_entry = ParsedNotification.objects.create(
-        maintenance=circuit_maintenance_entry,
-        raw_notification=raw_entry,
-        json=parsed_notification.to_json(),
-    )
-    logger.log_success(parsed_entry, message=f"Saved Parsed Notification for {maintenance_id}.")
+    return circuit_maintenance_entry
 
 
-def process_raw_notification(logger: Job, notification: MaintenanceNotification) -> Union[int, None]:
-    """Processes a raw notification (maybe containing multiple parsed notifications).
+def get_maintenances_from_notification(logger: Job, notification: MaintenanceNotification, provider: Provider):
+    """Use the `circuit_maintenance_parser` library to get Maintenances from the notification."""
+    provider_type = provider.cf.get("provider_parser_circuit_maintenances", "").lower() or provider.slug
 
-    It creates a RawNotification and if it could be parsed, create the corresponding ParsedNotification.
-    """
-    provider = Provider.objects.filter(slug=notification.provider_type).last()
-
-    if not provider:
-        logger.log_warning(
-            message=(
-                f"Raw notification not created because is referencing to a provider not existent: {notification.provider_type}"
-            )
-        )
+    parser_provider = init_provider(provider_type=provider_type)
+    if not parser_provider:
+        logger.log_warning(message=f"Notification Parser not found for {notification.provider_type}")
         return None
 
-    raw_payload = b""
-    for raw_payload in notification.raw_payloads:
-        parser = init_provider(raw=raw_payload, provider_type=notification.provider_type)
-        if not parser:
-            logger.log_warning(message=f"Notification Parser not found for {notification.provider_type}")
-            return None
+    data_to_process = NotificationData.init_from_email_bytes(notification.raw_payload)
+    if not data_to_process:
+        logger.log_failure(message=f"Notification data was not accepted by the parser: {notification.raw_payload}")
+        return None
 
-        try:
-            parsed_notifications = parser.process()
-            break
-        except ParsingError:
-            tb_str = traceback.format_exc()
-            logger.log_debug(message=f"Parsing failed for notification `{notification.subject}`:\n```\n{tb_str}\n```")
-        except Exception:
-            tb_str = traceback.format_exc()
-            logger.log_debug(
-                message=f"Unexpected exception while parsing notification `{notification.subject}`.\n```\n{tb_str}\n```"
-            )
-    else:
-        parsed_notifications = []
-        logger.log_warning(message=f"Parsed failed for all the raw payloads for `{notification.subject}`.")
-        # Carry on with the last raw_payload in the list, if any
-        if not raw_payload:
-            return None
+    try:
+        return parser_provider.get_maintenances(data_to_process)
+    except ProviderError:
+        tb_str = traceback.format_exc()
+        logger.log_failure(message=f"Parsing failed for notification `{notification.subject}`:\n```\n{tb_str}\n```")
+    except Exception:
+        tb_str = traceback.format_exc()
+        logger.log_failure(
+            message=f"Unexpected exception while parsing notification `{notification.subject}`.\n```\n{tb_str}\n```"
+        )
+    return None
 
-    if isinstance(raw_payload, str):
-        raw_payload = raw_payload.encode("utf-8")
 
+def create_raw_notification(logger: Job, notification: MaintenanceNotification, provider: Provider):
+    """Create a RawNotification.
+
+    If it already exists, we return `None` to signal we are skipping it.
+    """
     # Insert raw notification in DB even failed parsing
     try:
         raw_entry, created = RawNotification.objects.get_or_create(
             subject=notification.subject,
             provider=provider,
-            raw=raw_payload,
+            raw=notification.raw_payload,
             sender=notification.sender,
             source=NotificationSource.objects.filter(name=notification.source).last(),
         )
@@ -224,73 +212,130 @@ def process_raw_notification(logger: Job, notification: MaintenanceNotification)
         return None
 
     if not created:
-        logger.log_warning(message=f"Raw notification '{raw_entry.subject}' already existed with id {raw_entry.pk}")
+        # If the RawNotification was already created, we ignore it.
+        if logger.debug:
+            logger.log_debug(message=f"Raw notification '{raw_entry.subject}' already existed with id {raw_entry.pk}")
         return None
 
     logger.log_success(raw_entry, message="Raw notification created.")
 
-    for parsed_notification in parsed_notifications:
+    return raw_entry
+
+
+def process_raw_notification(logger: Job, notification: MaintenanceNotification) -> Optional[uuid.UUID]:
+    """Processes a raw notification (maybe containing multiple parsed notifications).
+
+    It creates a RawNotification and if it could be parsed, create the corresponding ParsedNotification and the
+    related objects. Finally returns the the UUID of the RawNotification modified.
+    """
+    provider = Provider.objects.filter(slug=notification.provider_type).last()
+    if not provider:
+        logger.log_warning(
+            message=(
+                f"Raw notification not created because is referencing to a provider not existent: {notification.provider_type}"
+            )
+        )
+        return None
+
+    raw_entry = create_raw_notification(logger, notification, provider)
+    if not raw_entry:
+        return None
+
+    parser_maintenances = get_maintenances_from_notification(logger, notification, provider)
+    if not parser_maintenances:
+        return raw_entry.id
+
+    for parser_maintenance in parser_maintenances:
         try:
-            process_parsed_notification(logger, parsed_notification, raw_entry)
-            # Update raw notification as properly parsed
+            circuit_maintenance_entry = create_or_update_circuit_maintenance(
+                logger, parser_maintenance, raw_entry, provider
+            )
+            # Update raw notification as properly parsed and with the stamp time
             raw_entry.parsed = True
+            raw_entry.date = datetime.datetime.fromtimestamp(parser_maintenance.stamp)
             raw_entry.save()
+
+            # Insert parsed notification in DB
+            parsed_entry = ParsedNotification.objects.create(
+                maintenance=circuit_maintenance_entry,
+                raw_notification=raw_entry,
+                json=parser_maintenance.to_json(),
+            )
+            logger.log_success(parsed_entry, message=f"Saved Parsed Notification for {circuit_maintenance_entry.name}.")
+
         except Exception as exc:
             tb_str = traceback.format_exception(etype=type(exc), value=exc, tb=exc.__traceback__)
-            logger.log_warning(
-                parsed_notification,
+            logger.log_failure(
+                parser_maintenance,
                 message=f"Unexpected exception while handling parsed notification {notification.subject}.\n{tb_str}",
             )
 
     return raw_entry.id
 
 
+def get_since_reference(logger: Job) -> int:
+    """Get the timestamp from the latest processed RawNotification or a reference from config `initial_days_since`."""
+    # Latest retrieved notification will limit the scope of notifications to retrieve
+    last_raw_notification = RawNotification.objects.last()
+    if last_raw_notification:
+        since_reference = last_raw_notification.date.timestamp()
+    else:
+        since_reference = datetime.datetime.utcnow() - datetime.timedelta(
+            days=PLUGIN_SETTINGS.get("raw_notifications", {}).get("initial_days_since", MAX_INITIAL_DAYS_SINCE)
+        )
+        since_reference = int(since_reference.timestamp())
+    logger.log_info(message=f"Processing notifications since {since_reference}")
+    return since_reference
+
+
 class HandleCircuitMaintenanceNotifications(Job):
     """Job to handle external circuit maintenance notifications and turn them into Circuit Maintenances."""
 
+    debug = BooleanVar(default=False, description="Enable getting debug Job messages.")
+
     class Meta:
         """Meta object boilerplate for HandleParsedNotifications."""
 
         name = "Update Circuit Maintenances"
-        description = (
-            "Fetch Circuit Maintenance Notifications from Source and create or update Circuit Maintenances accordingly"
-        )
+        description = "Fetch Circuit Maintenance Notifications from Sources and create or update Circuit Maintenances accordingly."
 
-    def run(self, data=None, commit=None):
+    def run(self, data=None, commit=None) -> List[uuid.UUID]:
         """Fetch notifications, process them and update Circuit Maintenance accordingly."""
-        self.log_debug("Starting Handle Notifications job.")
-        raw_notification_ids = []
+        if self.debug is True:
+            self.log_debug("Starting Handle Notifications job.")
+
         notification_sources = NotificationSource.objects.all()
         if not notification_sources:
             self.log_warning(message="No notification sources configured to retrieve notifications from.")
-            return raw_notification_ids
-
-        # Latest retrieved notification will limit the scope of notifications to retrieve
-        last_raw_notification = RawNotification.objects.last()
-        if last_raw_notification:
-            last_time_processed = last_raw_notification.date.timestamp()
-            self.log_info(message=f"Processing notifications since {last_raw_notification.date}")
-        else:
-            last_time_processed = None
+            return []
 
         try:
             notifications = get_notifications(
                 job_logger=self,
                 notification_sources=notification_sources,
-                since=last_time_processed,
+                since=get_since_reference(self),
             )
-            if not notifications:
-                self.log_info(message="No notifications received.")
-                return raw_notification_ids
+        except Exception as error:
+            self.log_failure(
+                message=f"Unexpected exception when retrieving notifications from sources ({notification_sources}): {error}"
+            )
+            return []
 
-            for notification in notifications:
-                self.log_info(message=f"Processing notification `{notification.subject}`.")
+        if not notifications:
+            self.log_info(message="No notifications received.")
+            return []
+
+        raw_notification_ids = []
+        for notification in notifications:
+            self.log_info(message=f"Processing notification `{notification.subject}`.")
+            try:
                 raw_id = process_raw_notification(self, notification)
                 if raw_id:
                     raw_notification_ids.append(raw_id)
+            except Exception as error:
+                self.log_failure(message=f"Unexpected exception when parsing notifications: {error}")
 
-        except Exception as error:
-            self.log_failure(message=f"Unexpected exception in Handle Notifications Job: {error}")
+        if self.debug is True:
+            self.log_debug(f"{len(raw_notification_ids)} notifications processed.")
 
-        self.log_debug(f"{len(raw_notification_ids)} notifications processed.")
         return raw_notification_ids
