@@ -8,8 +8,7 @@ import email
 import json
 import traceback
 from urllib.parse import urlparse
-from email.utils import mktime_tz, parsedate_tz
-from typing import Iterable, Optional, TypeVar, Type, Tuple, Dict, Union
+from typing import Iterable, List, Optional, TypeVar, Type, Tuple, Dict, Union
 
 import imaplib
 
@@ -23,10 +22,8 @@ from django.utils.text import slugify
 
 from pydantic import BaseModel  # pylint: disable=no-name-in-module
 from pydantic.error_wrappers import ValidationError  # pylint: disable=no-name-in-module
-from circuit_maintenance_parser import NonexistentParserError, get_provider_data_types
 from nautobot.circuits.models import Provider
 from nautobot.extras.jobs import Job
-from nautobot.extras.models import CustomField
 
 from nautobot_circuit_maintenance.models import NotificationSource
 
@@ -45,7 +42,7 @@ class MaintenanceNotification(BaseModel):
     sender: str
     subject: str
     provider_type: str
-    raw_payloads: Iterable[bytes]
+    raw_payload: bytes
 
 
 class Source(BaseModel):
@@ -58,15 +55,17 @@ class Source(BaseModel):
         """Method to get an identifier of the related account."""
         raise NotImplementedError
 
-    def receive_notifications(self, job_logger: Job, since: int = None) -> Iterable[MaintenanceNotification]:
+    def receive_notifications(
+        self, job_logger: Job, since_timestamp: datetime.datetime = None
+    ) -> Iterable[MaintenanceNotification]:
         """Function to retrieve notifications since one moment in time.
 
         The `MaintenanceNotification` attributes will contains these attributes:
         * source: self.name
         * sender: it could be the email 'from' or omitted if not relevant
         * subject: it could be the email 'subject' or some meaningful identifier from notification
+        * provider_type: mapping to the Provider that is related to this notification
         * raw: the raw_payload from notification
-        * provider_type: the provider slug identified by sender email or from membership
         """
         # TODO: `senders` is used to limit the scope of emails retrieved, this won't have sense depending on the
         # Notification Source.
@@ -162,6 +161,7 @@ class Source(BaseModel):
                     account=config.get("account"),
                     credentials_file=creds_filename,
                     source_header=config.get("source_header", "From"),
+                    extra_scopes=config.get("extra_scopes", []),
                 )
 
         raise ValueError(
@@ -200,8 +200,7 @@ class EmailSource(Source):  # pylint: disable=abstract-method
             return False
 
         for provider in notification_source.providers.all():
-            cm_cf = CustomField.objects.get(name="emails_circuit_maintenances")
-            provider_emails = provider.get_custom_fields().get(cm_cf)
+            provider_emails = provider.cf.get("emails_circuit_maintenances")
             if provider_emails:
                 self.emails_to_fetch.extend([src.strip().lower() for src in provider_emails.split(",")])
                 providers_with_email.append(provider.name)
@@ -223,8 +222,8 @@ class EmailSource(Source):  # pylint: disable=abstract-method
                 ),
             )
             return False
-
-        # job_logger.log_debug(message=f"Fetching emails from {self.emails_to_fetch}")
+        if job_logger.debug is True:
+            job_logger.log_debug(message=f"Fetching emails from {self.emails_to_fetch}")
         job_logger.log_info(
             message=(
                 f"Retrieving notifications from {notification_source.name} for "
@@ -247,39 +246,41 @@ class EmailSource(Source):  # pylint: disable=abstract-method
         return email_source.lower()
 
     @staticmethod
-    def extract_provider_data_types(email_source: str) -> Tuple[str, str, str]:
-        """Method to extract the provider data type based on the referenced email for the provider.
-
-        Returns:
-            Tuple(
-                Iterable[str]: provider_data_types, data types related to a specific Provider Parser
-                str: provider_type, corresponding to the Provider slug
-                str: error_message, if there was an issue
-            )
-        """
-        email_source = email_source.lower()
+    def get_provider_type_from_email(email_source: str) -> Optional[str]:
+        """Return the `Provider` type related to the source."""
         for provider in Provider.objects.all():
-            if not provider.custom_field_data.get("emails_circuit_maintenances"):
+            emails_for_provider = provider.cf.get("emails_circuit_maintenances")
+            if not emails_for_provider:
                 continue
-            sources = [
-                src.strip().lower() for src in provider.custom_field_data["emails_circuit_maintenances"].split(",")
-            ]
+            sources = [src.strip().lower() for src in emails_for_provider.split(",")]
             if email_source in sources:
-                provider_type = provider.slug
-                break
-        else:
-            return "", "", f"Sender email {email_source} is not registered for any circuit provider."
+                return provider.slug
+        return None
 
-        try:
-            provider_data_types = get_provider_data_types(provider_type)
-        except NonexistentParserError:
-            return (
-                "",
-                "",
-                f"{email_source} is for provider type `{provider_type}`, which is not presently supported",
+    def process_email(
+        self, job_logger: Job, email_message: email.message.EmailMessage
+    ) -> Optional[MaintenanceNotification]:
+        """Process an EmailMessage to create the MaintenaceNotification."""
+        email_source = self.extract_email_source(email_message[self.source_header])
+        if not email_source:
+            job_logger.log_warning(
+                message="Not possible to determine the email sender from "
+                f'"{self.source_header}: {email_message[self.source_header]}"'
             )
+            return None
 
-        return provider_data_types, provider_type, ""
+        provider_type = self.get_provider_type_from_email(email_source)
+        if not provider_type:
+            job_logger.log_warning(message=f"Not possible to determine the provider_type for {email_source}")
+            return None
+
+        return MaintenanceNotification(
+            source=self.name,
+            sender=email_source,
+            subject=email_message["Subject"],
+            provider_type=provider_type,
+            raw_payload=email_message.as_bytes(),
+        )
 
 
 class IMAP(EmailSource):
@@ -297,13 +298,20 @@ class IMAP(EmailSource):
         arbitrary_types_allowed = True
 
     def open_session(self):
-        """Open session to IMAP server."""
+        """Open session to IMAP server.
+
+        See states: https://github.com/python/cpython/blob/3.9/Lib/imaplib.py#L58
+        """
         if not self.session:
             self.session = imaplib.IMAP4_SSL(self.imap_server, self.imap_port)
+        if self.session.state == "NONAUTH":
             self.session.login(self.account, self.password)
 
     def close_session(self):
-        """Close session to IMAP server."""
+        """Close session to IMAP server.
+
+        See states: https://github.com/python/cpython/blob/3.9/Lib/imaplib.py#L58
+        """
         if self.session:
             if self.session.state == "SELECTED":
                 self.session.close()
@@ -315,62 +323,17 @@ class IMAP(EmailSource):
         self.open_session()
         self.close_session()
 
-    def fetch_email(self, job_logger: Job, msg_id: bytes, since: Optional[int]) -> Optional[MaintenanceNotification]:
+    def fetch_email(self, job_logger: Job, msg_id: bytes) -> Optional[MaintenanceNotification]:
         """Fetch an specific email ID."""
         _, data = self.session.fetch(msg_id, "(RFC822)")
         raw_email_string = data[0][1].decode("utf-8")
         email_message = email.message_from_string(raw_email_string)
 
-        if since:
-            if mktime_tz(parsedate_tz(email_message["Date"])) < since:
-                job_logger.log_info(message=f"'{email_message['Subject']}' email is old, so not taking into account.")
-                return None
-
         return self.process_email(job_logger, email_message)
 
-    def process_email(
-        self, job_logger: Job, email_message: email.message.EmailMessage
-    ) -> Optional[MaintenanceNotification]:
-        """Helper method for the fetch_email() method."""
-        email_source = self.extract_email_source(email_message[self.source_header])
-        if not email_source:
-            job_logger.log_warning(
-                message="Not possible to determine the email sender from "
-                f'"{self.source_header}: {email_message[self.source_header]}"'
-            )
-            return None
-
-        provider_data_types, provider_type, error_message = self.extract_provider_data_types(email_source)
-        if not provider_data_types:
-            job_logger.log_warning(message=error_message)
-            return None
-
-        raw_payloads = []
-        for provider_data_type in provider_data_types:
-            for part in email_message.walk():
-                if part.get_content_type() == provider_data_type:
-                    raw_payloads.append(part.get_payload())
-                    break
-            else:
-                job_logger.log_debug(
-                    message=f"Payload type {provider_data_type} not found in email payload.",
-                )
-
-        if not raw_payloads:
-            job_logger.log_warning(
-                message=f"Payload types {', '.join(provider_data_types)} not found in email payload.",
-            )
-            return None
-
-        return MaintenanceNotification(
-            source=self.name,
-            sender=email_message[self.source_header],
-            subject=email_message["Subject"],
-            raw_payloads=raw_payloads,
-            provider_type=provider_type,
-        )
-
-    def receive_notifications(self, job_logger: Job, since: int = None) -> Iterable[MaintenanceNotification]:
+    def receive_notifications(
+        self, job_logger: Job, since_timestamp: datetime.datetime = None
+    ) -> Iterable[MaintenanceNotification]:
         """Retrieve emails since an specific time, if provided."""
         self.open_session()
 
@@ -381,8 +344,8 @@ class IMAP(EmailSource):
         # Maybe extend filtering options, for instance, to discard some type of notifications
         msg_ids = []
         since_date = ""
-        if since:
-            since_txt = datetime.datetime.fromtimestamp(since).strftime("%d-%b-%Y")
+        if since_timestamp:
+            since_txt = since_timestamp.strftime("%d-%b-%Y")
             since_date = f'SINCE "{since_txt}"'
 
         if self.emails_to_fetch:
@@ -395,17 +358,34 @@ class IMAP(EmailSource):
                 search_criteria = f"({search_text})"
                 messages = self.session.search(None, search_criteria)[1][0]
                 msg_ids.extend(messages.split())
+                if job_logger.debug is True:
+                    job_logger.log_debug(
+                        message=(
+                            f"Fetched {len(messages.split())} emails from {self.name} source using search pattern: {search_criteria}."
+                        ),
+                    )
         else:
             search_criteria = f"({since_date})"
             messages = self.session.search(None, search_criteria)[1][0]
-            msg_ids.extend = messages.split()
+            msg_ids.extend(messages.split())
+            if job_logger.debug is True:
+                job_logger.log_debug(
+                    message=(
+                        f"Fetched {len(messages.split())} emails from {self.name} source using search pattern: {search_criteria}."
+                    ),
+                )
 
         received_notifications = []
         for msg_id in msg_ids:
-            raw_notification = self.fetch_email(job_logger, msg_id, since)
+            raw_notification = self.fetch_email(job_logger, msg_id)
             if raw_notification:
                 received_notifications.append(raw_notification)
 
+        if job_logger.debug:
+            job_logger.log_debug(
+                message=(f"Raw notifications created {len(received_notifications)} from {self.name}."),
+            )
+
         self.close_session()
         return received_notifications
 
@@ -420,6 +400,8 @@ class GmailAPI(EmailSource):
 
     SCOPES = ["https://www.googleapis.com/auth/gmail.readonly"]
 
+    extra_scopes: List[str] = []
+
     class Config:
         """Pydantic BaseModel config."""
 
@@ -458,95 +440,39 @@ class GmailAPI(EmailSource):
 
         return b""
 
-    def fetch_email(self, job_logger: Job, msg_id: bytes, since: Optional[int]) -> Optional[MaintenanceNotification]:
+    def fetch_email(self, job_logger: Job, msg_id: bytes) -> Optional[MaintenanceNotification]:
         """Fetch an specific email ID.
 
         See data format:  https://developers.google.com/gmail/api/reference/rest/v1/users.messages#Message
         """
         received_email = (
-            self.service.users().messages().get(userId=self.account, id=msg_id).execute()  # pylint: disable=no-member
+            self.service.users()  # pylint: disable=no-member
+            .messages()
+            .get(userId=self.account, id=msg_id, format="raw")
+            .execute()
         )
 
-        return self.process_email(job_logger, received_email, msg_id, since)
+        raw_email_string = base64.urlsafe_b64decode(received_email["raw"].encode("utf8"))
+        email_message = email.message_from_bytes(raw_email_string)
+        return self.process_email(job_logger, email_message)
 
-    def process_email(  # pylint: disable=too-many-locals
-        self, job_logger: Job, received_email: Dict, msg_id: bytes, since: Optional[int]
-    ) -> Optional[MaintenanceNotification]:
-        """Helper method for the fetch_email() method."""
-        email_subject = ""
-        email_source = ""
-
-        for header in received_email["payload"]["headers"]:
-            if header.get("name") == "Subject":
-                email_subject = header["value"]
-            elif header.get("name") == self.source_header:
-                email_source = header["value"]
-
-        if since:
-            if int(received_email["internalDate"]) < since:
-                job_logger.log_info(message=f"'{email_subject}' email is old, so not taking into account.")
-                return None
-
-        email_source_before = email_source
-        email_source = self.extract_email_source(email_source)
-        if not email_source:
-            job_logger.log_warning(message=f'Not possible to determine the email sender from "{email_source_before}"')
-            return None
-
-        provider_data_types, provider_type, error_message = self.extract_provider_data_types(email_source)
-        if not provider_data_types:
-            job_logger.log_warning(message=error_message)
-            return None
-
-        def get_raw_payload_from_parts(parts, provider_data_type):
-            """Helper function to extract the raw_payload from a multiple parts via a recursive call."""
-            for part_inner in parts:
-                for header_inner in part_inner["headers"]:
-                    if header_inner.get("name") == "Content-Type" and provider_data_type in header_inner.get("value"):
-                        return self.extract_raw_payload(part_inner["body"], msg_id)
-                    if header_inner.get("name") == "Content-Type" and "multipart" in header_inner.get("value"):
-                        return get_raw_payload_from_parts(part_inner["parts"], provider_data_type)
-            return None
-
-        raw_payloads = []
-        for provider_data_type in provider_data_types:
-            if "parts" not in received_email["payload"]:
-                raw_payload = self.extract_raw_payload(received_email["payload"]["body"], msg_id)
-            else:
-                raw_payload = get_raw_payload_from_parts(received_email["payload"]["parts"], provider_data_type)
-            if raw_payload:
-                raw_payloads.append(raw_payload)
-
-        if not raw_payloads:
-            job_logger.log_warning(
-                message=f"Payload types {provider_data_types} not found in email payload.",
-            )
-            return None
-
-        job_logger.log_debug(message=f"Got notification from {email_source} with subject {email_subject}")
-        return MaintenanceNotification(
-            source=self.name,
-            sender=email_source,
-            subject=email_subject,
-            raw_payloads=raw_payloads,
-            provider_type=provider_type,
-        )
-
-    def receive_notifications(self, job_logger: Job, since: int = None) -> Iterable[MaintenanceNotification]:
+    def receive_notifications(
+        self, job_logger: Job, since_timestamp: datetime.datetime = None
+    ) -> Iterable[MaintenanceNotification]:
         """Retrieve emails since an specific time, if provided."""
         self.load_credentials()
         self.build_service()
 
         search_criteria = ""
-        if since:
-            since_txt = datetime.datetime.fromtimestamp(since).strftime("%Y/%b/%d")
-            search_criteria = f'after:"{since_txt}"'
+        if since_timestamp:
+            since_txt = since_timestamp.strftime("%Y/%m/%d")
+            search_criteria = f"after:{since_txt}"
 
         # If source_header is not "from" but some other custom header such as X-Original-Sender,
         # the GMail API doesn't let us filter by that.
         if self.emails_to_fetch and self.source_header == "From":
             emails_with_from = [f"from:{email}" for email in self.emails_to_fetch]
-            search_criteria += f'({" OR ".join(emails_with_from)})'
+            search_criteria += " {" + f'{" ".join(emails_with_from)}' + "}"
 
         # TODO: For now not covering pagination as a way to limit the number of messages
         res = (
@@ -557,13 +483,24 @@ class GmailAPI(EmailSource):
         )
         msg_ids = [msg["id"] for msg in res.get("messages", [])]
 
+        if job_logger.debug is True:
+            job_logger.log_debug(
+                message=(
+                    f"Fetched {len(msg_ids)} emails from {self.name} source using search pattern: {search_criteria}."
+                ),
+            )
+
         received_notifications = []
         for msg_id in msg_ids:
-            raw_notification = self.fetch_email(job_logger, msg_id, since)
+            raw_notification = self.fetch_email(job_logger, msg_id)
             if raw_notification:
                 received_notifications.append(raw_notification)
 
-        # job_logger.log_debug(message=f"Raw notifications: {received_notifications}")
+        if job_logger.debug is True:
+            job_logger.log_debug(
+                message=(f"Raw notifications created {len(received_notifications)} from {self.name}."),
+            )
+            job_logger.log_debug(message=f"Raw notifications: {received_notifications}")
 
         self.close_service()
         return received_notifications
@@ -612,23 +549,21 @@ class GmailAPIServiceAccount(GmailAPI):
         """Load Gmail API Service Account credentials."""
         if not self.credentials:
             self.credentials = service_account.Credentials.from_service_account_file(self.credentials_file)
-            self.credentials = self.credentials.with_scopes(self.SCOPES)
+            self.credentials = self.credentials.with_scopes(self.SCOPES + self.extra_scopes)
             self.credentials = self.credentials.with_subject(self.account)
             self.credentials.refresh(Request())
 
 
 def get_notifications(
-    job_logger: Job, notification_sources: Iterable[NotificationSource], since: int = None
+    job_logger: Job, notification_sources: Iterable[NotificationSource], since: int
 ) -> Iterable[MaintenanceNotification]:
     """Method to fetch notifications from multiple sources and return MaintenanceNotification objects."""
     received_notifications = []
 
     for notification_source in notification_sources:
         try:
-            if since:
-                since_txt = datetime.datetime.fromtimestamp(since).strftime("%d-%b-%Y")
-            else:
-                since_txt = "always"
+            since_date = datetime.datetime.fromtimestamp(since)
+            since_txt = since_date.strftime("%d-%b-%Y")
 
             try:
                 source = Source.init(name=notification_source.name)
@@ -645,7 +580,12 @@ def get_notifications(
                 continue
 
             if source.validate_providers(job_logger, notification_source, since_txt):
-                raw_notifications = source.receive_notifications(job_logger, since)
+                if since_date:
+                    # When using the SINCE filter, we add one extra day to check for notifications received
+                    # on the very same day since last notification.
+                    since_date -= datetime.timedelta(days=1)
+
+                raw_notifications = source.receive_notifications(job_logger, since_date)
                 received_notifications.extend(raw_notifications)
 
                 if not raw_notifications:
